(****************************************************************************)
(*                                                                          *)
(* Ce fichier contient les fonctions ncessaires  la cration des Grids et *)
(* et HexaGrid de 2 et 3 dimensions.                                        *)
(*                                                                          *)
(* La tche essentielle consiste  dterminer la nombre de colonnes, lignes *)
(* et tages en fonction des ratios.                                        *)
(*                                                                          *)
(****************************************************************************)



open Types



(**************************************************************)
(* Dtermine les 2 axes, le nombre de colonnes et de lignes   *)
(* d'un Grid 2 dimensions  partir des 2 axes entrs axis1 et *)
(* axis2, du ratio rate et du nombre d'lments n du Grid.    *)
(**************************************************************)
(* Type : axe -> axe -> float * float * float -> int ->       *)
(*        axe * axe * int * int                               *)
(**************************************************************)
let axis_ratio2 axis1 axis2 rate n =
  let axis11 = ref axis1
  and axis22 = ref axis2
  in
    (* On dtermine les axes en fonction des valeurs entres.                              *)
    (* Si aucun axe n'est entr, les objets seront rangs suivant l'axe X puis l'axe Z.    *)
    (* Si un seul axe est dtermin ou si les 2 axes fix sont gaux, une erreur est lev. *)
    if (axis1 = AxeNone && axis2 = AxeNone)
    then
      begin
	axis11 := AxeX;
	axis22 := AxeZ;
      end
    else if (axis1 = AxeNone && axis2 <> AxeNone) || (axis1 <> AxeNone && axis2 = AxeNone)
    then raise (ParseError "un axe du grid2 n'est pas dfini");

    let (r1, r2) = let (rt1, rt2, rt3) = rate
    in
      (match (!axis11, !axis22)
       with
	 | (AxeX, AxeY) | (AxeY, AxeX) -> (rt1, rt2)
	 | (AxeY, AxeZ) | (AxeZ, AxeY) -> (rt2, rt3)
	 | (AxeZ, AxeX) | (AxeX, AxeZ) -> (rt3, rt1)
	 | (   _,    _)                -> raise (ParseError "Cration du grid : repre invalide")
      )
    in
      (* On vrifie que les valeur entres pour le ratio ne sont strictement positives *)
      if (r1 <= 0.0) || (r2 <= 0.0)
      then raise (ParseError "Cration du grid : ratio invalide");

      (* On calcule un nombre de colonnes et de lignes tels que le ratio *)
      (* obtenu soit le plus proche possible du ratio souhait.          *)
      let n_f = float_of_int n
      in
      let tmp = sqrt (n_f /. (r1 *. r2))
      in
      let rr1 = ref (int_of_float (r1 *. tmp))
      and rr2 = ref (int_of_float (r2 *. tmp))
      in
	if (!rr1 * !rr2 < n)
	then
	  begin
	    incr rr2;
	    if (!rr1 * !rr2 < n)
	    then
	      begin
		incr rr1;
		rr2 := int_of_float (ceil (n_f /. (float_of_int !rr1)));
	      end
	  end;
	(!axis11, !axis22, !rr1, !rr2)



(***********************************************************)
(* Dtermine les 2 premiers axes, le nombre de colonnes et *)
(* de lignes et d'tages d'un Grid 3 dimensions  partir   *)
(* des 2 axes entrs axis1 et axis2, du ratio (r1, r2, r3) *)
(* et du nombre d'lments n du Grid.                      *)
(***********************************************************)
(* Type : axe -> axe -> float -> float -> float -> int ->  *)
(*        axe * axe * int * int * int                      *)
(***********************************************************)
let axis_ratio3 axis1 axis2 r1 r2 r3 n =
  let axis11 = ref axis1
  and axis22 = ref axis2
  in
    (* On dtermine les axes en fonction des valeurs entres.                              *)
    (* Si aucun axe n'est entr, les objets seront rangs suivant l'axe X puis l'axe Y.    *)
    (* Si un seul axe est dtermin ou si les 2 axes fix sont gaux, une erreur est lev. *)
    if (axis1 = AxeNone && axis2 = AxeNone)
    then
      begin
	axis11 := AxeX;
	axis22 := AxeY;
      end
    else if (axis1 = AxeNone && axis2 <> AxeNone) || (axis1 <> AxeNone && axis2 = AxeNone)
	 then raise (ParseError "un axe du grid3 n'est pas dfini")
	 else if axis1 = axis2
	      then raise (ParseError "Cration du grid : repre invalide");

   (* On vrifie que les valeur entres pour le ratio ne sont strictement positives *)
   if (r1 <= 0.0) || (r2 <= 0.0) || (r3 <= 0.0)
   then raise (ParseError "Cration du grid : ratio invalide");

   (* On calcule un nombre de colonnes et de lignes tels que le ratio *)
   (* obtenu soit le plus proche possible du ratio souhait.          *)
   let n_f = float_of_int n
   in
   let tmp = (n_f /. (r1 *. r2 *. r3)) ** (0.333333333333333333333333333333333333333333333333333333333)
   in
   let rr1 = ref (int_of_float (r1 *. tmp))
   and rr2 = ref (int_of_float (r2 *. tmp))
   and rr3 = ref (int_of_float (r3 *. tmp))
   in
     if (!rr1 * !rr2 * !rr3 < n)
     then
       begin
	 incr rr3;
	 if (!rr1 * !rr2 * !rr3 < n)
	 then
	   begin
	     incr rr2;
	     rr3 := min !rr3 (int_of_float (ceil (n_f /. (float_of_int (!rr1 * !rr2)))));
	     if (!rr1 * !rr2 * !rr3 < n)
	     then
	       begin
		 incr rr1;
		 decr rr2;
		 rr3 := min !rr3 (int_of_float (ceil (n_f /. (float_of_int (!rr1 * !rr2)))));
		 if (!rr1 * !rr2 * !rr3 < n)
		 then
		   begin
		     incr rr2;
		     rr3 := int_of_float (ceil (n_f /. (float_of_int (!rr1 * !rr2))));
		   end
	       end
	   end
       end;
     (!axis11, !axis22, !rr1, !rr2, !rr3)



(************************************************************)
(* Cration d'un Grid2                                      *)
(*   pt     : centre                                        *)
(*   axis1  : 1er axe de remplissage                        *)
(*   axis2  : 2me axe de remplissage                       *)
(*   align1 : alignement sur le 1er axe                     *)
(*   align2 : alignement sur le 2me axe                    *)
(*   align3 : alignement sur le 3me axe                    *)
(*   space1 : espace minimum entre 2 objets sur le 1er axe  *)
(*   space2 : espace minimum entre 2 objets sur le 2me axe *)
(*   rate   : triplet entr pour le ratio                   *)
(*   coul   : couleur du Grid2                              *)
(*   lst    : liste des objets du Grid2                     *)
(************************************************************)
(* Type : point -> axe -> axe -> align -> align -> align -> *)
(*        float -> float -> float * float * float ->        *)
(*        couleur -> (objet * point) list -> objet          *)
(************************************************************)
let create_grid2 (pt, axis1, axis2, align1, align2, align3, space1, space2, rate, coul, lst) =
  let (axis11, axis22, rr1, rr2) = axis_ratio2 axis1 axis2 rate (List.length lst)
  in
    Grid (pt, Grid2 (axis11, axis22, align1, align2, align3, space1, space2, rr1, rr2), coul, lst)



(************************************************************)
(* Cration d'un Grid2                                      *)
(*   pt     : centre                                        *)
(*   axis1  : 1er axe de remplissage                        *)
(*   axis2  : 2me axe de remplissage                       *)
(*   align1 : alignement sur le 1er axe                     *)
(*   align2 : alignement sur le 2me axe                    *)
(*   align3 : alignement sur le 3me axe                    *)
(*   space1 : espace minimum entre 2 objets sur le 1er axe  *)
(*   space2 : espace minimum entre 2 objets sur le 2me axe *)
(*   space3 : espace minimum entre 2 objets sur le 3me axe *)
(*   r1     : valeur entre pour le ratio sur l'axe axis1   *)
(*   r2     : valeur entre pour le ratio sur l'axe axis2   *)
(*   r3     : valeur entre pour le ratio sur l'axe "axis3" *)
(*   coul   : couleur du Grid2                              *)
(*   lst    : liste des objets du Grid2                     *)
(************************************************************)
(* Type : point -> axe -> axe -> align -> align -> align -> *)
(*        float -> float -> float -> float -> float         *)
(*        float -> couleur -> (objet * point) list -> objet *)
(************************************************************)
let create_grid3 (pt, axis1, axis2, align1, align2, align3, space1, space2, space3, r1, r2, r3, coul, lst) =
  let (axis11, axis22, rr1, rr2, rr3) = axis_ratio3 axis1 axis2 r1 r2 r3 (List.length lst)
  in
    Grid (pt, Grid3 (axis11, axis22, align1, align2, align3, space1, space2, space3, rr1, rr2, rr3), coul, lst)



(************************************************************)
(* Cration d'un HexaGrid2                                  *)
(*   pt     : centre                                        *)
(*   axis1  : 1er axe de remplissage                        *)
(*   axis2  : 2me axe de remplissage                       *)
(*   align1 : alignement sur le 1er axe                     *)
(*   align2 : alignement sur le 2me axe                    *)
(*   align3 : alignement sur le 3me axe                    *)
(*   space  : espace entre 2 objets                         *)
(*   rate   : triplet entr pour le ratio                   *)
(*   coul   : couleur de l'HexaGrid2                        *)
(*   lst    : liste des objets de l'HexaGrid2               *)
(************************************************************)
(* Type : point -> axe -> axe -> align -> align -> align -> *)
(*        float -> float * float * float -> couleur ->      *)
(*        (objet * point) list -> objet                     *)
(************************************************************)
let create_hexa2 (pt, axis1, axis2, align3, space, rate, coul, lst) =
  let (axis11, axis22, rr1, rr2) = axis_ratio2 axis1 axis2 rate (List.length lst)
  in
    Grid (pt, Hexa2 (axis11, axis22, align3, space, rr1, rr2), coul, lst)



(************************************************************)
(* Cration d'un Grid2                                      *)
(*   pt     : centre                                        *)
(*   axis1  : 1er axe de remplissage                        *)
(*   axis2  : 2me axe de remplissage                       *)
(*   align1 : alignement sur le 1er axe                     *)
(*   align2 : alignement sur le 2me axe                    *)
(*   align3 : alignement sur le 3me axe                    *)
(*   space  : espace entre 2 objets                         *)
(*   r1     : valeur entre pour le ratio sur l'axe axis1   *)
(*   r2     : valeur entre pour le ratio sur l'axe axis2   *)
(*   r3     : valeur entre pour le ratio sur l'axe "axis3" *)
(*   coul   : couleur du Grid2                              *)
(*   lst    : liste des objets du Grid2                     *)
(************************************************************)
(* Type : point -> axe -> axe -> align -> align -> align -> *)
(*        float -> float -> float -> float -> couleur ->    *)
(*        (objet * point) list -> objet                     *)
(************************************************************)
let create_hexa3 (pt, axis1, axis2, space, r1, r2, r3, coul, lst) =
  let (axis11, axis22, rr1, rr2, rr3) = axis_ratio3 axis1 axis2 r1 r2 r3 (List.length lst)
  in
    Grid (pt, Hexa3 (axis11, axis22, space, rr1, rr2, rr3), coul, lst)
