(***************************************************************************)
(*                                                                         *)
(* Ce fichier contient les fonctions ncessaires  l'affichage des scnes. *)
(*                                                                         *)
(***************************************************************************)



open Types
open Triplet



let x = [| true ;  true ; false ; false ;  true ;  true ; false ;
	   true ;  true ; false ; false ;  true ;  true ; false ;
	   true ;  true ; false ; false ;  true ;  true ; false ;
	  false ;  true ;  true ;  true ;  true ; false ; false ;
	  false ; false ;  true ;  true ; false ; false ; false ;
	  false ;  true ;  true ;  true ;  true ; false ; false ;
	   true ;  true ; false ; false ;  true ;  true ; false ;
	   true ;  true ; false ; false ;  true ;  true ; false ;
	   true ;  true ; false ; false ;  true ;  true ; false
	|]

let y = [| false ; false ;  true ;  true ; false ; false ; false ;
	   false ; false ;  true ;  true ; false ; false ; false ;
	   false ; false ;  true ;  true ; false ; false ; false ;
	   false ; false ;  true ;  true ; false ; false ; false ;
	   false ;  true ;  true ;  true ;  true ; false ; false ;
	    true ;  true ; false ; false ;  true ;  true ; false ;
	    true ;  true ; false ; false ;  true ;  true ; false ;
	    true ;  true ; false ; false ;  true ;  true ; false ;
	    true ;  true ; false ; false ;  true ;  true ; false
	|]

let z = [| true ;  true ;  true ;  true ;  true ;  true ;  true ;
	   true ;  true ; false ; false ; false ;  true ;  true ;
	   true ;  true ; false ; false ; false ; false ;  true ;
	  false ;  true ;  true ; false ; false ; false ; false ;
	  false ; false ;  true ;  true ; false ; false ; false ;
	  false ;  true ; false ;  true ;  true ; false ; false ;
	   true ; false ; false ; false ;  true ;  true ; false ;
	   true ;  true ; false ; false ;  true ;  true ;  true ;
	   true ;  true ;  true ;  true ;  true ;  true ;  true
	|]

let plus = [| false ; false ; false ; false ; false ; false ; false ;
	      false ; false ; false ; false ; false ; false ; false ;
	      false ; false ;  true ;  true ; false ; false ; false ;
	      false ; false ;  true ;  true ; false ; false ; false ;
	       true ;  true ;  true ;  true ;  true ;  true ; false ;
	      false ; false ;  true ;  true ; false ; false ; false ;
	      false ; false ;  true ;  true ; false ; false ; false ;
	      false ; false ; false ; false ; false ; false ; false ;
	      false ; false ; false ; false ; false ; false ; false
	   |]

let moins = [| false ; false ; false ; false ; false ; false ; false ;
	       false ; false ; false ; false ; false ; false ; false ;
	       false ; false ; false ; false ; false ; false ; false ;
	       false ; false ; false ; false ; false ; false ; false ;
	        true ;  true ;  true ;  true ;  true ;  true ; false ;
	       false ; false ; false ; false ; false ; false ; false ;
	       false ; false ; false ; false ; false ; false ; false ;
	       false ; false ; false ; false ; false ; false ; false ;
	       false ; false ; false ; false ; false ; false ; false
	   |]



(******************************************************************)
(* Fusionne le signe et le nom et transforme le tout en un Raw.t  *)
(* de couleurs.                                                   *)
(******************************************************************)
(* Type : bool array -> bool array -> float array -> `float Raw.t *)
(******************************************************************)
let mat2pix sign axe color = let raw = Raw.create `float ~len:504
			     in
			       for i = 0 to 6
			       do
				 for j = 0 to 8
				 do
				   let charSeek = 7*j+i
				   and  rawSeek = (14*j+i)*4
				   in
				     if sign.(charSeek)
				     then
				       Raw.sets_float raw ~pos:rawSeek color
				     else
				       Raw.sets_float raw ~pos:rawSeek [| 0. ; 0. ; 0. ; 0. |];
				     if axe.(charSeek)
				     then
				       Raw.sets_float raw ~pos:(rawSeek+28) color
				     else
				       Raw.sets_float raw ~pos:(rawSeek+28) [| 0. ; 0. ; 0. ; 0. |]
				 done
			       done;
			       raw



(*****************************)
(* Pixmap des noms des axes. *)
(*****************************)
let plusX = ((GlPix.of_raw (mat2pix plus x [| 1. ; 0. ; 0. ; 1. |])
			   ~format:`rgba
			   ~width:14
			   ~height:9
	     ):([`rgba], [`float]) GlPix.t)

let moinsX = ((GlPix.of_raw (mat2pix moins x [| 1. ; 0. ; 0. ; 1. |])
			    ~format:`rgba
			    ~width:14
			    ~height:9
	      ):([`rgba], [`float]) GlPix.t)

let plusY = ((GlPix.of_raw (mat2pix plus y [| 0. ; 1. ; 0. ; 1. |])
			   ~format:`rgba
			   ~width:14
			   ~height:9
	     ):([`rgba], [`float]) GlPix.t)

let moinsY = ((GlPix.of_raw (mat2pix moins y [| 0. ; 1. ; 0. ; 1. |])
			    ~format:`rgba
			    ~width:14
			    ~height:9
	      ):([`rgba], [`float]) GlPix.t)
let plusZ = ((GlPix.of_raw (mat2pix plus z [| 0. ; 0. ; 1. ; 1. |])
			   ~format:`rgba
			   ~width:14
			   ~height:9
	     ):([`rgba], [`float]) GlPix.t)

let moinsZ = ((GlPix.of_raw (mat2pix moins z [| 0. ; 0. ; 1. ; 1. |])
			    ~format:`rgba
			    ~width:14
			    ~height:9
	      ):([`rgba], [`float]) GlPix.t)



(**********************************)
(* Liste des objets transparents. *)
(**********************************)
let blendObj = (ref []:((Types.primitive * Types.couleur * Types.draw) * float * GlMat.t) list ref)



(******************************************************************)
(* Insre dans lst un objet suivant sa profondeur z dcroissante. *)
(* lst est une liste de mme sorte que blendObj.                  *)
(******************************************************************)
(* Type : (Types.primitive * Types.couleur * Types.draw) as `a -> *)
(*        float -> GlMat.t -> (`a * float * GlMat.t) list ->      *)
(*        (objet * float * GlMat.t) list                          *)
(******************************************************************)
let rec insertOrdonne obj z mat lst = match lst
				      with
					| []   -> (obj, z, mat)::[]
					| t::q -> let (_, tZ, _) = t
						  in
						    if tZ >= z
						    then
						      t::(insertOrdonne obj z mat q)
						    else
						      (obj, z, mat)::lst



(***********************************************)
(* Fixe la couleur de l'objet  partir des     *)
(* composantes rouge, vert, bleu et alpha      *)
(***********************************************)
(* Type : float * float *float * float -> unit *)
(***********************************************)
let color (rouge, vert, bleu, alpha) = GlDraw.color ~alpha:alpha (rouge, vert, bleu);
				       GlLight.material `both (`diffuse (rouge, vert, bleu, alpha));
				       GlLight.material `both (`ambient (rouge, vert, bleu, alpha))



(*****************************************)
(* Dessine les axes de longueur 2*taille *)
(*****************************************)
(* Type : float -> unit                  *)
(*****************************************)
let draw_axes taille =
  (* axe X *)
  GlDraw.color (1.0, 0.0, 0.0);
  GlDraw.begins `lines;
  GlDraw.vertex3 (-. taille,       0.0,       0.0);
  GlDraw.vertex3 (   taille,       0.0,       0.0);
  (* axe Y *)
  GlDraw.color (0.0, 1.0, 0.0);
  GlDraw.vertex3 (      0.0, -. taille,       0.0);
  GlDraw.vertex3 (      0.0,    taille,       0.0);
  (* axe Z *)
  GlDraw.color (0.0, 0.0, 1.0);
  GlDraw.vertex3 (      0.0,       0.0, -. taille);
  GlDraw.vertex3 (      0.0,       0.0,    taille);
  GlDraw.ends ()



(*****************************************)
(* Dessine le nom des axes de longueur.  *)
(* taille est la demi-longueur des axes. *)
(*****************************************)
(* Type : float -> unit                  *)
(*****************************************)
let draw_axes_name taille = Gl.enable `blend;
			    GlPix.raster_pos ~x:taille ~y:1.0 ~z:0.0 ();
			    GlPix.draw plusX;

			    GlPix.raster_pos ~x:(-.taille) ~y:1.0 ~z:0.0 ();
			    GlPix.draw moinsX;

			    GlPix.raster_pos ~x:1.0 ~y:taille ~z:0.0 ();
			    GlPix.draw plusY;

			    GlPix.raster_pos ~x:1.0 ~y:(-.taille) ~z:0.0 ();
			    GlPix.draw moinsY;

			    GlPix.raster_pos ~x:0.0 ~y:1.0 ~z:taille ();
			    GlPix.draw plusZ;

			    GlPix.raster_pos ~x:0.0 ~y:1.0 ~z:(-.taille) ();
			    GlPix.draw moinsZ;

			    Gl.disable `blend



(*********************************)
(* Dessine la grille sur le plan *)
(* form des axes axis1 et axis2 *)
(*********************************)
(* Type : axe -> axe -> unit     *)
(*********************************)
let draw_grille axis1 axis2 =
  let limite = floor !Parametres.rayon
  in
  let i = ref (-.limite)
  in
  let axis_func = match (axis1, axis2)
		  with
		    | (AxeX, AxeY) | (AxeY, AxeX) -> (function i -> GlDraw.vertex3 (-. !Parametres.rayon,                   i,                 0.0);
								    GlDraw.vertex3 (   !Parametres.rayon,                   i,                 0.0);
								    GlDraw.vertex3 (                   i,-. !Parametres.rayon,                 0.0);
								    GlDraw.vertex3 (                   i,   !Parametres.rayon,                 0.0))
		    | (AxeY, AxeZ) | (AxeZ, AxeY) -> (function i -> GlDraw.vertex3 (                 0.0,-. !Parametres.rayon,                   i);
								    GlDraw.vertex3 (                 0.0,   !Parametres.rayon,                   i);
								    GlDraw.vertex3 (                 0.0,                   i,-. !Parametres.rayon);
								    GlDraw.vertex3 (                 0.0,                   i,   !Parametres.rayon))
		    | (AxeZ, AxeX) | (AxeX, AxeZ) -> (function i -> GlDraw.vertex3 (                   i,                 0.0,-. !Parametres.rayon);
								    GlDraw.vertex3 (                   i,                 0.0,   !Parametres.rayon);
								    GlDraw.vertex3 (-. !Parametres.rayon,                 0.0,                   i);
								    GlDraw.vertex3 (   !Parametres.rayon,                 0.0,                   i))
		    | _                            -> ignore
  in
    GlDraw.color (0.7, 0.7, 0.7);
    GlDraw.begins `lines;

    while !i < limite
    do
      axis_func !i;
      i := !i +. 1.0
    done;
    GlDraw.ends ()



(***********************************************)
(* Dessine une primitive sans son contenu.     *)
(* drawBlend indique si un objet transparent   *)
(* doit tre affich ou inscrit dans blendObj. *)
(* secteur indique quel morceau afficher :     *)
(*   - Inf    -> 1/2 sphre infrieur          *)
(*   - Sup    -> 1/2 sphre suprieur          *)
(*   - All    -> sphre entire                *)
(*   - Bottom -> Ple Sud                      *)
(***********************************************)
(* Type : Types.primitive -> Types.couleur ->  *)
(*        bool -> Types.draw -> unit           *)
(***********************************************)
let rec drawPrim (prim, coul, wire) drawBlend secteur =
  match wire
  with
    | Wire when !Parametres.polygon_mode = `fill ->
	color (coul#getRouge, coul#getVert, coul#getBleu, coul#getAlpha);
	GlDraw.polygon_mode ~face:`both `line;
	GlDraw.cull_face `back;
	Gl.disable `cull_face;
	Primitives.draw_primitive prim secteur false;
	GlDraw.cull_face `back;
	Gl.enable `cull_face;
	GlDraw.polygon_mode ~face:`both !Parametres.polygon_mode
	  
    | Mixed (down, up) ->
	drawPrim (prim, coul, down) drawBlend Inf;
	drawPrim (prim, coul, up  ) drawBlend Sup
	  
    | Open ->
	color (coul#getRouge, coul#getVert, coul#getBleu, coul#getAlpha);
	Primitives.draw_primitive prim Bottom true
	  
    | Wire
    | Close ->
	color (coul#getRouge, coul#getVert, coul#getBleu, coul#getAlpha);
	Primitives.draw_primitive prim secteur (secteur <> All || coul#getAlpha <> 1.)
	  
    | Blend alpha ->
	if drawBlend
	then
	  begin
	    color (coul#getRouge, coul#getVert, coul#getBleu, coul#getAlpha *. alpha);
	    Primitives.draw_primitive prim secteur true
	  end
	else
	  begin
	    let mat = GlMat.get_matrix `modelview_matrix
	    and newWire =
	      match secteur with
		| Bottom -> (* Impossible :@ *) Shadow
		| Inf    -> Mixed (Blend alpha,      Shadow)
		| Sup    -> Mixed (     Shadow, Blend alpha)
		| All    -> Blend alpha
	    in
	      blendObj := insertOrdonne (prim, coul, newWire) (Raw.get_float (GlMat.to_raw mat) ~pos:11) mat !blendObj
	  end

    | Shadow -> ()




(************************)
(* Dessine un objet obj *)
(************************)
(* Type : objet -> unit *)
(************************)
and draw_objet obj =
  match obj with
    | Vide -> ()
    | Primitive (prim, coul, wire, cont) ->
	(match cont with
	   | None             -> ()
	   | Some (c, h, lst) ->
	       GlMat.push ();
	       if !h <> 1.0
	       then
		 GlMat.scale3 (!h, !h, !h);
	       GlMat.translate3 c#getAll;
	       List.iter draw_objet lst;
	       GlMat.pop ()
	);
	if coul#getAlpha <> 1.
	then
	  begin
	    let mat = GlMat.get_matrix `modelview_matrix
	    in
	      blendObj := insertOrdonne (prim, coul, !wire) (Raw.get_float (GlMat.to_raw mat) ~pos:11) mat !blendObj
	  end
	else
	  drawPrim (prim, coul, !wire) false All
      
    | Transformation (trans, objTrans) ->
	GlMat.push ();
	(match trans with
	   | Rotation (axis, angle) ->
	       GlMat.rotate3 angle (match axis
				    with
				      | AxeX    -> (1., 0., 0.)
				      | AxeY    -> (0., 1., 0.)
				      | AxeZ    -> (0., 0., 1.)
				      | AxeNone -> raise (Erreur "Draw : axe de rotation invalide")
				   )
	   | Scale pt -> GlMat.scale3 pt#getAll
	   | Translation pt -> GlMat.translate3 pt#getAll
	);
	draw_objet objTrans;
	GlMat.pop ()

    | Groupe (_, objGrp) -> List.iter draw_objet objGrp
	
    | Grid (pt, _, _, objList) ->
	GlMat.push ();
	GlMat.translate3 pt#getAll;
	List.iter (function (obj, pt) -> GlMat.translate3 pt#getAll; draw_objet obj) objList;
	GlMat.pop ()



(****************************************)
(* Dessine toutes les scnes depuis une *)
(* camera excentre de l'axe de delta   *)
(****************************************)
(* Type : float -> unit                 *)
(****************************************)
let view delta =
  (* On met  jour la matrice de projection *)
  Move.set_perspective ();
  GlMat.load_identity ();
  (* On place la lumire *)
  GlLight.light ~num:0 (`position (*10.0, 10.0, 10.0, 0.0*)
			(*let (lposX,lposY,lposZ) = add_triplet
						    (add_triplet
						       Parametres.centre#getAll
						       (mult_triplet
							  (Parametres.vectZ#getAll)
							  !Parametres.zoom
						       )
						    )
						    (mult_triplet Parametres.vectX#getAll delta)
			in
			  (lposX, lposY, lposZ, 0.0)*)
			(0.0, 0.0, !Parametres.zoom, 0.0)
		       );

  (* On place la camera *)
  GluMat.look_at ~eye:(add_triplet
			 (add_triplet
			    Parametres.centre#getAll
			    (mult_triplet
			       (Parametres.vectZ#getAll)
			       !Parametres.zoom
			    )
			 )
			 (mult_triplet Parametres.vectX#getAll delta)
		      )
		 ~center:Parametres.centre#getAll
		 ~up:Parametres.vectY#getAll;

  (* Rotation de la scne sur elle-mme *)
  GlMat.translate3 Parametres.milieu_dessin#getAll;
  GlMat.mult (GlMat.of_array !Parametres.spin);
  GlMat.translate3 (mult_triplet Parametres.milieu_dessin#getAll (-1.0));

  (* Affichage de la grille et des axes si ncessaire *)
  draw_grille !Parametres.grille1 !Parametres.grille2;
  if !Parametres.affiche_axe
  then draw_axes !Parametres.rayon;

  (* Acitivation du brouillard et de la lumire si ncessaire pour les scnes *)
  if !Parametres.lighting
  then Gl.enable `lighting;
  if !Parametres.brouillard
  then Gl.enable `fog;

  (* Dessin des scnes *)
  List.iter (fun scn -> if (scn#isShown)
			then List.iter draw_objet scn#getContenu
			else ()
	    )
	    !Parametres.dessin;

  (* Dessine les objets transparents *)
  let sbt () = GlFunc.depth_mask false;
		List.iter (fun (obj, _, mat) -> GlMat.load mat;
						drawPrim obj true All
			  )
			  !blendObj;
		GlFunc.depth_mask true
   in
     if !Parametres.isVectoring
     then
       begin
	 Vectoriel.psEnable `blend;
	 sbt ();
	 Vectoriel.psDisable `blend
       end
     else
       begin
	 Gl.enable `blend;
	 sbt ();
	 Gl.disable `blend
       end;
    blendObj := [];

  if !Parametres.affiche_axe
  then draw_axes_name !Parametres.rayon;

  Gl.disable `lighting;
  Gl.disable `fog



(****************************************)
(* Dessine toutes les scenes suivant si *)
(* le mode stro est activ ou non.    *)
(****************************************)
(* Type : unit -> unit                  *)
(****************************************)
let draw_scene () =
  if !Parametres.stereo
  then
    begin
      (*Mode stro *)
      GlFunc.draw_buffer `back;
      GlClear.clear [`color; `depth];
      (* Oeil droit *)
      GlFunc.draw_buffer `back_right;
      view !Parametres.inter_yeux;
      (* Oeil gauche *)
      GlFunc.draw_buffer `back_left;
      GlClear.clear [`depth];
      view (-. !Parametres.inter_yeux)
    end
  else
    begin
      (* Mode "mono" *)
      GlClear.clear [`color; `depth];
      view 0.0
    end;
  !Parametres.swapbuffers ()
