(***************************************************************************)
(*                                                                         *)
(* Ce fichier contient une classe permettant de choisir un nom de fichier. *)
(* Une fentre est ouverte et l'utilisateur dtermine un chemin et un nom  *)
(* pour le fichier.                                                        *)
(*                                                                         *)
(***************************************************************************)



(****************************************************************)
(* Type dterminant quelle action sera ralise sur le fichier. *)
(****************************************************************)
type fileAction =
  | Load (* Charge un fichier existant *)
  | Save (* Enregistre dans un fichier *)



(*****************************************************)
(* Classe ouvrant la fentre de slection du fichier *)
(*****************************************************)
(* Type : fileAction -> (string -> ()) -> class      *)
(*****************************************************)
class imo_fichier action path storeName =

  let title = "Imoview - " ^ match action
			     with
			       | Load -> "Ouvrir"
			       | Save -> "Enregistrer"
  in

  let window = GWindow.file_selection ~wm_name:title
				      ~title:title
				      ~position:`NONE
				      ~modal:true
				      ~allow_shrink:false
				      ~allow_grow:true
				      ~auto_shrink:false
				      ~border_width:10
				      ~fileop_buttons:true  
				      ()
  in

  let ok     = window#ok_button
  and cancel = window#cancel_button
  in
  let _ = GtkBase.Widget.set_can_focus   ok#as_widget true
  and _ = GtkBase.Widget.set_can_default ok#as_widget true
  and _ = GtkBase.Widget.set_can_focus   cancel#as_widget true
  and _ = GtkBase.Widget.set_can_default cancel#as_widget true
  in


object(self)



  method window = window
  method ok     = ok
  method cancel = cancel

  method on_quitter () = window#destroy ()
  method on_window_remove (widget:GObj.widget)  = self#on_quitter ()

  method valide () = storeName window#get_filename;
		     self#on_quitter ()



initializer 



  let _ = window#event#connect#key_press ~callback:(fun (event: GdkEvent.Key.t) -> begin
										     let key = GdkEvent.Key.keyval event
										     in
										       match key
										       with
											 | t when t=GdkKeysyms._Escape -> self#on_quitter ();
											     true;
											 | _                           -> false
										   end
						   )
  and _ = window#complete ~filter:path
  and _ = window#connect#remove   ~callback:self#on_window_remove
  and _ = ok    #connect#clicked  ~callback:self#valide
  and _ = cancel#connect#clicked  ~callback:self#on_quitter
  in
    ()

end
