open Types
open GMain
open Parametres



(*************************************)
(* Gestion des option du programme  *)
(* partir du tableau d'arguments arg *)
(*************************************)
(* Type : string array -> unit       *)
(*************************************)
let options arg =
  Arg.parse [ "-axis", Arg.Set affiche_axe, "\tactive les axes" ;

	      "-fontdpi", Arg.String (function str -> defaultDpi := (int_of_string str)), "ttf\trsoluation pour l'affichage des textes [defaut=72]" ;

	      "-file", Arg.String (function str -> if (!entree = File) && (!file_name = "")
						   then file_name := str
						   else
						     begin
						       prerr_string ("Option \"-file " ^ str ^ "\" invalide\n");
						       exit 1
						     end
				  ), "name\tla lecture se fait depuis le fichier name" ;

	      "-gridXY", Arg.Unit (function () -> grille1 := AxeX;
						  grille2 := AxeY
				  ), "\taffiche la grille sur le plan XY" ;

	      "-gridYZ", Arg.Unit (function () -> grille1 := AxeY;
						  grille2 := AxeZ
				  ), "\taffiche la grille sur le plan YZ" ;

	      "-gridZX", Arg.Unit (function () -> grille1 := AxeZ;
						  grille2 := AxeX
				  ), "\taffiche la grille sur le plan ZX" ;

	      "-noaxis", Arg.Clear affiche_axe, "\tdsactive les axes" ;

	      "-nogrid", Arg.Unit (function () -> grille1 := AxeNone;
						  grille2 := AxeNone
				  ), "\tdsactive la grille" ;

	      "-nofog", Arg.Clear brouillard, "\tdsactive la brouillard" ;

	      "-nolight", Arg.Clear lighting, "\tdsactive la lumire" ;

	      "-size [x:y]", Arg.String (function str -> let dims = Str.split (Str.regexp ":") (String.sub str 1 ((String.length str) - 2))
						   in
						     try
						       width  := float_of_string (List.hd dims);
						       height := float_of_string (List.hd (List.tl dims));
						     with
							 _ -> prerr_endline "Erreur dans l'option -size";
							      exit 1
				  ), "\tdtermine la taille de la zone d'affichage 3D";

	      "-stereo", Arg.Set stereo, "\tmode stro" ;

	      "-stdin", Arg.Unit (function () -> if (!entree = File) && (!file_name <> "")
						 then
						   begin
						     prerr_string ("Argument " ^ !file_name ^ " invalide ou incompatible avec l'option -stdin\n");
						     exit 1
						   end
						 else
						   begin
						     entree    := Stdin;
						     file_name := "imoview"
						   end
				 ), "\tla lecture se fait depuis l'entre standard" ;


	      "-font", Arg.String (function str -> defaultTtf := str), "ttf\tl'affichage des textes utilise ttf comme police par dfaut" ;

	      "-withZbug", Arg.Set withZbug, "\tprend l'oppos de la cte (Z) pour les polylignes et les polygones"

	    ]
	    (function str -> if (!entree = File) && (!file_name = "")
			     then
			       begin
				 entree    := File;
				 file_name := str
			       end
			     else
			       begin
				 prerr_string ("Argument " ^ str ^ " invalide\n");
				 exit 1
			       end
	    )
	    ("Usage : " ^ arg.(0) ^ " [options] [fichier]\n")



(***********************)
(* Fonction principale *)
(***********************)
let main () =
  let arguments = Sys.argv
  in
  let argc = Array.length arguments
  in
    (* Gestion des options de lancement *)
    options arguments;

    Load.load_file ();
    let window = new Guimain.imo_window ()
    in
    let _ = window#window#show ()
    in
      GMain.Main.main ()

let _ = Printexc.print main ()
