(********************************************************************)
(*                                                                  *)
(* Le fichier load.ml contient la fonction pour charger un fichier. *)
(*                                                                  *)
(********************************************************************)



open Types
open Parametres



(******************************)
(* Mise  zro des paramtres *)
(* du parser et d'affichage.  *)
(******************************)
(* Type : unit -> unit        *)
(******************************)
let init_parser () = dico           := [];
		     dessin         := [];
		     sceneLst       := [];
		     select         := Vide;
		     line           := 1;
		     command_todo#clear ();
		     command_done#clear ();

		     vectX #setAll (1., 0.              , 0.              );
		     vectY #setAll (0., 0.86602540378445,-0.5             );
		     vectZ #setAll (0., 0.5             , 0.86602540378445);
		     centre#setAll (0., 0.              , 0.              )



(****************************************)
(* Analyse l' inchannel (~flux) chan    *)
(****************************************)
(* Type : inchannel -> unit             *)
(****************************************)
let analyse win chan = try
			 let lexBuffer = Lexing.from_channel chan
			 in
			   (* Mise  "zro" des paramtres de l'analyse syntaxique *)
			   init_parser ();
			   (* Analyse syntaxique *)
			   if !Parametres.entree = Stdin
			   then
			     begin
			       while (Parser.std  Scanner.token lexBuffer)
			       do
				 ()
			       done;
			       Draw.draw_scene ();
			     end
			   else Parser.file Scanner.token lexBuffer

		       (* On rattrape les erreurs syntaxiques pour indiquer le type d'erreur et la localisation *)
		       with
			 | ParseError s -> close_in chan;
					   My_controls.messageBox "Erreur dans le chargement" (" ligne " ^ (string_of_int !line) ^ " : " ^ s)



(*********************************************)
(* Analyse et lancer l'affichage  partir de *)
(* l'entre dfinie dans Parametres.entree   *)
(*********************************************)
(* Type : unit -> unit                       *)
(*********************************************)
let load_file () = match (!entree, !file_name)
		   with
		     (* On analyse le contenu de l'entre standard *)
		     | (Stdin, _)                     -> () (*ignore (Thread.create (analyse (Glut.getWindow ())) stdin)*)

		     (* On analyse le fichier name *)
		     | (File, name) when (name <> "") -> (try
							    let chan = open_in name
							    in
							      analyse 0 chan;
							      (* On ferme le fichier *)
							      close_in chan

							  (* On rattrape les erreurs systmes (impossible d'ouvrir le fichier ...) *)
							  with
							    | Sys_error s -> My_controls.messageBox "Erreur dans le chargement" s
							 )

		     (* Nom de fichier vide -> on met tout  vide *)
		     | (File, _)                      -> init_parser ()
