(********************************************************************)
(*                                                                  *)
(* Ce fichier contient les fonctions ncessaires au mouvements lis *)
(*  la manipulation de la souris et  la commande de mouvement.    *)
(*                                                                  *)
(********************************************************************)



open Types
open Triplet



(***************************************)
(* Redfinit la matrice de perspective *)
(***************************************)
(* Type : unit -> unit                 *)
(***************************************)
let set_perspective () =
  let w = !Parametres.width
  and h = !Parametres.height
  and distance = !Parametres.zoom +. (scalaire Parametres.vectZ#getAll (diff_triplet Parametres.centre#getAll Parametres.milieu_dessin#getAll))
  and r = !Parametres.rayon *. Parametres.sqrt2
  in
  let near = max 5.0 (distance -. r)
  in
    GlMat.mode `projection;
    GlMat.load_identity ();
    GluMat.perspective ~fovy:45.0 ~aspect:(w/.h) ~z:(near, distance +. r);

    GlMat.mode `modelview



(*****************************************************)
(* Mise  jour de v1 et v2 aprs la rotation d'angle *)
(* theta autour de v1 vectoriel v2                   *)
(*****************************************************)
(* Type : float * float * float ->                   *)
(*        float * float * float -> unit              *)
(*****************************************************)
let rotate_vect v1 v2 theta =
  let coord1 = v1#getAll
  and coord2 = v2#getAll
  and cos_theta = cos theta
  and sin_theta = sin theta
  in
    v1#setAll (add_triplet (mult_triplet coord1 cos_theta) (mult_triplet coord2 sin_theta));
    v2#setAll (add_triplet (mult_triplet coord2 cos_theta) (mult_triplet coord1 (-.sin_theta)))



(****************************************************)
(* Multiplie 2 matrices 4*4 a et b.                 *)
(****************************************************)
(* Type : float array array -> float array array -> *)
(*        float array array                         *)
(****************************************************)
let mult_matrice a b = let res = [| Array.make 4 0.;
				    Array.make 4 0.;
				    Array.make 4 0.;
				    Array.make 4 0.
				 |]
		       in
			 for i = 0 to 3
			 do
			   for j = 0 to 3
			   do
			     res.(i).(j) <- (a.(i).(0) *. b.(0).(j)) +. (a.(i).(1) *. b.(1).(j)) +. (a.(i).(2) *. b.(2).(j))+. (a.(i).(3) *. b.(3).(j))
			   done
			 done;
			 res



(**********************************************************)
(* Multipline la matrice spin (fichier parametres.ml) par *)
(* la matrice de rotation d'angle theta et d'axe vect     *)
(* angle est en degrs.                                   *)
(**********************************************************)
(* Type : float * float * float -> float -> unit          *)
(**********************************************************)
let mult_spin vect theta = let angle = theta *. Parametres.pi /. 180.0
			   in
			   let c = cos angle
			   and s = sin angle
			   and (x, y, z) = normalise_triplet vect
			   in
			     Parametres.spin := mult_matrice !Parametres.spin [| [| x*.x*.(1.-.c)+.c    ; x*.y*.(1.-.c)-.z*.s ; x*.z*.(1.-.c)+.y*.s ; 0. |];
										 [| y*.x*.(1.-.c)+.z*.s ; y*.y*.(1.-.c)+.c    ; y*.z*.(1.-.c)-.x*.s ; 0. |];
										 [| x*.z*.(1.-.c)-.y*.s ; y*.z*.(1.-.c)+.x*.s ; z*.z*.(1.-.c)+.c    ; 0. |];
										 [| 0.                  ; 0.                  ; 0.                  ; 1. |]
									      |]



(*****************************************************)
(* Rotation de la camera autour du centre de l'cran *)
(* x, y, z sont les angles en degrs de la rotation  *)
(* autour des axes du repre de la camera            *)
(*****************************************************)
(* Type : float * float * float -> unit              *)
(*****************************************************)
let rotation (x, y, z) = if x <> 0.0
			 then rotate_vect Parametres.vectY Parametres.vectZ (Parametres.pi *. x /. 180.);
			 if y <> 0.0
			 then rotate_vect Parametres.vectZ Parametres.vectX (Parametres.pi *. y /. 180.);
			 if z <> 0.0
			 then rotate_vect Parametres.vectX Parametres.vectY (Parametres.pi *. z /. 180.)



(****************************************************)
(* Rotation du dessin sur lui-mme                  *)
(* x, y, z sont les angles en degrs de la rotation *)
(* autour des axes du repre de la camera.          *)
(* x0, y0, z0 sont les angles en degrs de la       *)
(* rotation autour des axes du repre absolu.       *)
(****************************************************)
(* Type : float * float * float ->                  *)
(*        float * float * float -> unit             *)
(****************************************************)
let spin (x, y, z) (x0, y0, z0) =
  if  x <> 0.0
  then mult_spin Parametres.vectX#getAll     x;
  if  y <> 0.0
  then mult_spin Parametres.vectY#getAll     y;
  if  z <> 0.0
  then mult_spin Parametres.vectZ#getAll     z;
  if x0 <> 0.0
  then mult_spin (1.0, 0.0, 0.0) x0;
  if y0 <> 0.0
  then mult_spin (0.0, 1.0, 0.0) y0;
  if z0 <> 0.0
  then mult_spin (0.0, 0.0, 1.0) z0



(*************************************************)
(* Translation de la camera de vecteur (x, y, z) *)
(* dans le repre de la camera                   *)
(*************************************************)
(* Type : float * float * float -> unit          *)
(*************************************************)
let translation (x, y, z) =
  let trans = add_triplet (mult_triplet (Parametres.vectX#getAll) x) (add_triplet (mult_triplet (Parametres.vectY#getAll) y) (mult_triplet (Parametres.vectZ#getAll) z))
  in
    Parametres.centre#updateAll trans
