class iconLabeled obj (label:string) (icon:GDraw.pixmap) =

  let ico = GMisc.pixmap icon
			 ~xalign:0.5
			 ~yalign:0.5
			 ~xpad:0
			 ~ypad:0
			 ()
  in

  let lbl = GMisc.label ~text:label
			~xalign:0.5
			~yalign:0.5
			~xpad:0
			~ypad:0
			~line_wrap:false
			()
  in

object(self)

  inherit GPack.box obj

  method text = lbl#text
  method setText str = lbl#set_text str
  method changeIcone newIcon = ico#set_pixmap newIcon
  method icon = ico
  method label = lbl

initializer

  let _ = GObj.pack_return self#icon
			   ~packing:(Some (self#pack ~padding:0
						     ~fill:false
						     ~expand:false
					  )
				    )
			   ~show:(Some true)

  and _ = GObj.pack_return self#label
			   ~packing:(Some (self#pack ~padding:0
						     ~fill:false
						     ~expand:false
					  )
				    )
			   ~show:(Some true)

  in
    ()

end



let iconLabeled label icon ?packing ?show () = let w = GtkPack.Box.create `HORIZONTAL
									  ~spacing:5
									  ~homogeneous:false
									  ()
					       in
						 GObj.pack_return (new iconLabeled w label icon)
								  ~packing
								  ~show




class toolButton obj (icon:GDraw.pixmap) =

  let button = GButton.button ()
  in

  let _ = GtkBase.Widget.set_can_focus button#as_widget true
  in

  let ico = GMisc.pixmap icon
			 ~xalign:0.5
			 ~yalign:0.5
			 ~xpad:0
			 ~ypad:0
			 ()
  in

object(self)

  inherit GButton.button obj

  method changeIcone newIcon = ico#set_pixmap newIcon
  method icon = ico

initializer

  let _ = GObj.pack_return self#icon
			   ~packing:(Some (self#add))
			   ~show:(Some true)

  in
    ()

end



let toolButton icon ?tips ?packing ?show () = let w = GtkButton.Button.create ()
					      in
					      let control = new toolButton w icon
					      in
						(match tips
						 with
						   | None -> ()
						   | Some f -> f control#coerce
						);
						GObj.pack_return (new toolButton w icon)
								 ~packing
								 ~show




let messageBox titre message = 

  let window = GWindow.window ~wm_name:titre
			      ~title:("Imoview - " ^ titre)
			      ~position:`CENTER
			      ~kind:`DIALOG
			      ~modal:true
			      ~allow_shrink:false
			      ~allow_grow:false
			      ~auto_shrink:false
			      ()
  in

  let vbox = GPack.vbox ~spacing:10
			~homogeneous:false
			~packing:window#add
			()
  in

  let hbox = GPack.hbox ~spacing:10
			~homogeneous:false
			~packing:(vbox#pack ~fill:true
					    ~expand:true
				 )
			()
  in

  let ico = GMisc.pixmap (GDraw.pixmap_from_xpm_d ~data:Icones.warning
						  ~window:window
						  ()
			 )
			 ~xalign:0.5
			 ~yalign:0.5
			 ~xpad:10
			 ~ypad:10
			 ~packing:(hbox#pack ~fill:true
					     ~expand:true
				  )
			 ()
  in

  let lbl = GMisc.label ~text:message
			~xalign:0.5
			~yalign:0.5
			~xpad:10
			~ypad:10
			~line_wrap:true
			 ~packing:(hbox#pack ~fill:true
					     ~expand:true
				  )
			()
  in

  let bottom = GPack.button_box `HORIZONTAL
				~spacing:30
				~child_width:85
				~child_height:27
				~layout:`DEFAULT_STYLE
				~packing:(vbox#pack ~padding:10
						    ~fill:true
						    ~expand:true
					 )
				()
  in

  let retour = GButton.button ~label:"Retour"
			      ~packing:(bottom#pack)
			      ()
  in

  let _ = GtkBase.Widget.set_can_focus   retour#as_widget true
  and _ = GtkBase.Widget.set_can_default retour#as_widget true
  and _ = retour#grab_default ()
  in

  let on_quitter () = window#destroy ()
  in

    (* Gestion du clavier *)
    ignore (window#event#connect#key_press ~callback:(fun (event: GdkEvent.Key.t) -> begin
										       let key = GdkEvent.Key.keyval event
										       in
											 match key
											 with
											   | t when t=GdkKeysyms._Escape -> on_quitter ();
															    true
											   | _                           -> false
										     end
						     )
	   );
    ignore (window#connect#remove  ~callback:(fun (widget:GObj.widget) -> on_quitter ()));
    ignore (retour#connect#clicked ~callback:on_quitter);

    window#show ()
