(***************************************************************************)
(*                                                                         *)
(* Le fichier parametres.ml contient toutes sortes de variables "globlaes" *)
(* utilises dans plusieurs fichiers diffrents.                           *)
(*                                                                         *)
(***************************************************************************)



open Types
open File



(**************)
(* Constantes *)
(**************)


(* Pi *)
let pi = 3.1415926535897932384626433832795029

(*  2 Pi  *)
let pi2  = 6.2831853071795864769252867665590058

(* Pi / 2 *)
let pi_2 = 1.5707963267948966192313216916397514

(* Sin(Pi/8) *)
let sin_pi_8 = 0.3826834323651

(* racine de 2 *)
let sqrt2 = 1.4142135623730950488016887242096981

(*       Racine de 3        *)
let sqrt3 = 1.732050807568877

(*       Racine de 6        *)
let sqrt6 = 2.449489742783178



(***********************************)
(* Donnes de l'analyse syntaxique *)
(***********************************)

(* Nom du fichier source *)
let file_name = ref ""

(* Nom du fichier source *)
let entree = ref File


(* Dictionnaire du parser *)
let dico = ref ([]:((string * objet) list))


(* Dictionnaire des scenes *)
let sceneLst = ref ([]:((string * scene) list))


(* Numro de ligne de la lecture du fichier *)
let line = ref 1


(*************************************)
(* Donnes pour l'affichage de texte *)
(*************************************)

let defaultTtf = ref ""

let defaultDpi = ref 72

(*****************)
(* Axes, grilles *)
(*****************)


(* Affichage ou non des axes *)
let affiche_axe    = ref false


(* Axes formants le plan de la grille *)
(* Un plan invalide provode le        *)
(* non-affichage de la grille         *)
let grille1 = ref AxeNone
let grille2 = ref AxeNone



(********************)
(* Mode d'affichage *)
(********************)


(* Type d'affichage des polygones  (Fill, Line ou Point) *)
let polygon_mode = ref (`fill:[`fill | `line | `point])


(* Eclair ou non *)
let lighting = ref true

(* Brouillard ou non *)
let brouillard = ref true

(* Stereo ou non *)
let stereo = ref false

(* Bug de l'oppos du Z *)
let withZbug = ref false

(* Ecartement des yeux *)
let inter_yeux = ref 1.0



(****************)
(* Point de vue *)
(****************)


(* Vecteurs orientant la vue :
	vectX : tel que (vectX, vectY, vectZ) orthonorme
	vectY : dfinie la verticale
	vectZ : vecteur unitaire porte par l'axe optique
 *)
let vectX = new point 1.0 0.0 0.0
let vectY = new point 0.0 1.0 0.0
let vectZ = new point 0.0 0.0 1.0


(* Point objerv *)
let centre = new point 0.0 0.0 0.0


(* Milieu du dessin *)
let milieu_dessin = new point 0.0 0.0 0.0


(* Focus : distance entre la camera et le point observ *)
let zoom = ref 10.0


(* Matrice de rotation du dessin sur lui-mme *)
let spin = ref ([| [| 1.0 ; 0.0 ; 0.0 ; 0.0 |];
		   [| 0.0 ; 1.0 ; 0.0 ; 0.0 |];
		   [| 0.0 ; 0.0 ; 1.0 ; 0.0 |];
		   [| 0.0 ; 0.0 ; 0.0 ; 1.0 |]
	       |])



(**********)
(* Dessin *)
(**********)


(* Liste des scenes *)
let dessin  = ref ([]:(scene list))

(* Liste des scnes affiches aprs la dernire commande *)
let lstShownEnd = ref ([]:(scene list))


(* File des commandes  raliser *)
let command_todo = (new file:(commande file))


(* File des commandes dj ralises *)
let command_done = (new file:(commande file))


(* Taille du dessin *)
let rayon = ref 0.0


(* Etat de l'animation *)
let state = ref Stop


(* Etat de l'animation *)
let nextState = ref (None:(anim * sens) option)


(* Timer de l'animation *)
let timer = ref (None:GMain.Timeout.id option)

(* Intervale de temps entre 2 comandes *)
let cmdTimout = ref 1


(* Direction de la lecture *)
let sens = ref Avant



(****************************)
(* Slection des Containers *)
(****************************)


(* Container slectionn *)
let select = ref Vide


(* Etat du clignotement (taux de transparence) *)
let clignote_rate = ref 1.0


(* ~drive de clignote_rate *)
let clignote_sign = ref 0.1



(*********************************)
(* Etat des boutons de la souris *)
(*********************************)


(* Bouton gauche  de la souris *)
let   leftButton = new mouseButton


(* Bouton central de la souris *)
let middleButton = new mouseButton


(* Bouton droit   de la souris *)
let  rightButton = new mouseButton



(*******************)
(* Capture d'cran *)
(*******************)


(* indice du screenshot *)
let shot = ref 0


(* Indique si une capture dans un format vectoriel est en cours *)
let isVectoring = ref false




(*******************)
(* Var pour guigtk *)
(*******************)


  (* swap-buffers de l'aire OpenGL *)
  let swapbuffers = ref (fun () -> ())

  (* GTKGL area *)
  let area = ref (None:GlGtk.area option)

  (* taille de l'aire OpenGL *)
  let width = ref 500.0
  let height = ref 400.0

  let ecrireStatut = ref (fun (s:string)->())

(*  let idle = ref(GMain.Timeout.add ~ms:100000 ~callback:(fun ()->true))*)

  let encodeLine = ref "transcode --use_rgb -i %list -x imlist,null -g %wx%h -y xvid,null -f %fps -o %nom -H 0"
