(* TODO : commenter le fichier *)
(*
open Parametres
open Triplet
open Types



(*********************************************************************)
(* Dtermine l'indice du container obj dans la liste des containers  *)
(* Les indices commencent  1.                                       *)
(* Retourne 0 si obj n'est pas dans la liste des containers.         *)
(*********************************************************************)
(* Type : objet -> int                                               *)
(*********************************************************************)
let container2int obj =
  let rec _container2int obj lst i = match lst
				     with
				       | hd::tl when (hd = obj) -> i
				       | _::tl                  -> _container2int obj tl (i+1)
				       | []                     -> 0
  in
    _container2int obj !Parametres.container_list 1



(**************************************************************)
(* Dtermine le container d'indice i la liste des containers. *)
(* Les indices commencent  1.                                *)
(* Retourne Vide si i est un indice invalide.                 *)
(**************************************************************)
(* Type : int -> objet                                        *)
(**************************************************************)
let int2container i = if (i <= 0) || (i > (List.length !Parametres.container_list))
		      then Vide
		      else List.nth !Parametres.container_list (i-1)



let rec draw_picking_objet obj = match obj
				 with
				   | Vide                                -> ()

				   | Primitive (prim, coul, _, cont)     -> Draw.color (coul#getRouge, coul#getVert, coul#getBleu, coul#getAlpha);
									    Primitives.draw_primitive prim

				   | Transformation (trans, objTrans)    -> GlMat.push ();
									    (match trans
									     with
									       | Rotation (axis, angle) -> GlMat.rotate3 angle (match axis
																with
																  | AxeX    -> (1.0, 0.0, 0.0)
																  | AxeY    -> (0.0, 1.0, 0.0)
																  | AxeZ    -> (0.0, 0.0, 1.0)
																  | AxeNone -> raise (Erreur "Picking : axe de rotaion invalide")
															       )
									       | Scale pt               -> GlMat.scale3 pt#getAll
									       | Translation pt         -> GlMat.translate3 pt#getAll
									    );
									    draw_picking_objet objTrans;
									    GlMat.pop ()

				   | Groupe (_, objGrp)                  -> List.iter draw_picking_objet objGrp

				   | Container (typ, drw, coul, center, objCont) -> GlMat.push ();
										    GlMat.translate3 center#getAll;
										    GlMat.pop ();
										    List.iter draw_picking_objet objCont;
										    GlMisc.load_name (container2int obj);
							 			    Draw.draw_container typ (match !drw
													     with
													       | Blend  -> Close
													       | Half   -> Half
													       | Mixed  -> Close
													       | Shadow -> Shadow
													       | Open   -> Open
													       | Wire   -> Wire
													       | Close  -> Close
													    ) coul 1.0

				   | Grid (pt, _, _, objList)            -> GlMat.push ();
									    GlMat.translate3 pt#getAll;
									    List.iter (function (obj, pt) -> GlMat.translate3 pt#getAll;
													     GlMat.push ();
													     draw_picking_objet obj;
													     GlMat.pop ()
										      ) objList;
									    GlMat.pop ()



let picking x y =
  let h = !(Parametres.height)
  and w = !(Parametres.width)
  and n = List.length !Parametres.container_list + 1
  in
  let buffer = Raw.create_static `uint ~len:(4 * n)
  in
    for i=0 to (4*n-1) do Raw.set buffer ~pos:i 0 done;
    GlMisc.select_buffer buffer;
    ignore (GlMisc.render_mode `select);
    GlMisc.init_names ();
    GlMisc.push_name 0;

    (* Matrice de projection *)
    GlMat.mode `projection;
    GlMat.push ();
    GlMat.load_identity ();
    GluMat.pick_matrix ~x:x ~y:(h -. y) ~width:2.0 ~height:2.0;
    let distance = !zoom +. (scalaire vectZ#getAll (diff_triplet centre#getAll milieu_dessin#getAll))
    and r = !rayon *. sqrt2
    in
    let near = max 5.0 (distance -. r)
    in
      GluMat.perspective ~fovy:45.0 ~aspect:(w/.h) ~z:(near, distance +. r);

      (* Matrice de vue *)
      GlMat.mode `modelview;
      GlMat.load_identity ();
      GluMat.look_at ~eye:(add_triplet centre#getAll (mult_triplet (vectZ#getAll) !zoom))
		     ~center:centre#getAll
		     ~up:vectY#getAll;

      GlMat.translate3 milieu_dessin#getAll;
      GlMat.mult (GlMat.of_array !spin);
      GlMat.translate3 (mult_triplet milieu_dessin#getAll (-1.0));

      (* Dessin des objets *)
      List.iter (function scn -> if (scn#isShown)
				 then List.iter draw_picking_objet scn#getContenu
				 else ()
		) !dessin;

      (* On remets la matrice de projection en tat *)
      GlMat.mode `projection;
      GlMat.pop ();
      GlMat.mode `modelview;

      let hits = GlMisc.render_mode `render
      and depth  = ref (Raw.get buffer ~pos:2)
      and choose = ref (Raw.get buffer ~pos:3)
      in
      let plop = Raw.gets_string buffer ~pos:0 ~len:(4*n* (Raw.sizeof `uint))
      in 
	if hits > 0
	then
	  begin
	    for i = 1 to (hits - 1)
	    do
	      if ((Raw.get buffer ~pos:((i*4)+1)) < !depth)
	      then
		begin
		  depth  := (Raw.get buffer ~pos:((i*4)+1));
		  choose := (Raw.get buffer ~pos:((i*4)+3))
		end
	    done;
	    Raw.free_static buffer;
	    int2container !choose
	  end
	else Vide



let selectionne x y =
  let new_select = picking x y
  in
    if new_select <> Vide
    then
      begin
	if !select = new_select
	then
	  begin
	    select := Vide;
            GtkMain.Timeout.remove (!Parametres.idle);
	    Draw.draw_scene ();
	  end
	else
	  begin
	    select := new_select;
	    clignote_rate := 1.0;
	    clignote_sign := 0.1;
            Draw.draw_scene ();
            GtkMain.Timeout.remove (!Parametres.idle);
            Parametres.idle := GtkMain.Timeout.add ~ms:20 ~callback:(function () -> clignote_rate := !clignote_rate +. !clignote_sign;
										    (if (!clignote_rate > 1.5)
										     then
										       begin
											 clignote_rate :=  1.0;
											 clignote_sign := -0.1
										       end
										     else
										       begin
											 if (!clignote_rate < -0.5)
											 then
											   begin
											     clignote_rate := 0.0;
											     clignote_sign := 0.1
											   end
										       end
										    );
										    Draw.draw_scene ();
										    true
								    )
	  end
      end
*)
