(**********************************************)
(*                                            *)
(* Ce fichier contient la fonction ncessaire *)
(* pour raliser une capture d'cran.         *)
(*                                            *)
(**********************************************)



(*******************************************************)
(* Renvoie une capture d'cran sous forme matricielle. *)
(*******************************************************)
(* Type : float -> Rgb24.t                             *)
(*******************************************************)
let saveImage () = let w = int_of_float !(Parametres.width)
		   and h = int_of_float  !(Parametres.height)
		   in
		   let pixmap = GlPix.read ~x:0 ~y:0 ~width:w ~height:h ~format:`rgb ~kind:`ubyte
		   in
		   let raw    = GlPix.to_raw pixmap
		   and image = Rgb24.create w h
		   in
		     (* On enregistre dans image chaque pixel de la fentre OpenGL *)
		     for i = 0 to w - 1 do
		       for j = 0 to h - 1 do
			 match (Raw.gets raw ~pos:(3*(j*w+i)) ~len:3)
			 with
			   | [|r;g;b|] -> Rgb24.set image i (h-1-j) {Color.r = r; Color.g = g; Color.b = b}
			   | _         -> ()
		       done
		     done;
		     image



let enregImage nom format options img rate = let small = if rate = 1.0
							 then
							   img
							 else
							   Rgb24.resize None
									img
									(Usefull.round (!Parametres.width  *. rate))
									(Usefull.round (!Parametres.height *. rate))
					     in
					       (match format
						with
						  | Image.Gif -> let index8 = (Coupe_mediane.coupeMediane small)
								 in
								   print_string "Avant save\n"; flush stdout;
								   Printf.printf "%d %d\n" index8.Index8.width index8.Index8.height; flush stdout;
								   Image.save nom (Some format) options (Image.Index8 index8);
								   print_string "Apres save\n"; flush stdout;
								   Index8.destroy index8;
								   print_string "Apres destroy\n"; flush stdout

						  | _         -> Image.save nom (Some format) options (Image.Rgb24  small)
					       );
					       if rate <> 1.0
					       then
						 Rgb24.destroy small



(******************************************************************)
(* Enregistre une capture du contenu de la fentre OpenGL sous    *)
(* forme vectorielle dans un fichier de nom file.                 *)
(* Le format de l'image dpend de format.                         *)
(* background indique si l'arrire plan doit tre incorpor  la  *)
(* capture.                                                       *)
(* titre sera enregistr dans les paramtres de l'image.          *)
(* buferSize indique la taille du buffer  utiliser.              *)
(******************************************************************)
(* Type : string -> Types.format -> bool -> string -> int -> unit *)
(******************************************************************)
let saveVectSize titre format background file bufferSize =
  Vectoriel.psBeginPage titre format background bufferSize file;

  Vectoriel.psBlendFunc `src_alpha `one_minus_src_alpha;
  Vectoriel.psPointSize 2.0;
  Vectoriel.psLineWidth 1.0;

  Draw.draw_scene ();

  Vectoriel.psEndPage ()



(******************************************************************)
(* Enregistre une capture du contenu de la fentre OpenGL sous    *)
(* forme vectorielle dans un fichier de nom file.                 *)
(* Le format de l'image dpend de format.                         *)
(* background indique si l'arrire plan doit tre incorpor  la  *)
(* capture.                                                       *)
(* titre sera enregistr dans les paramtres de l'image.          *)
(******************************************************************)
(* Type : string -> Types.format -> bool -> string -> unit *)
(******************************************************************)
let saveVect titre format background file = let rec saveRec size = try
								     saveVectSize titre format background file size
								   with
								     | Failure s  -> if s = "Overflow"
										     then
										       saveRec (2*size)
										     else
										       begin
											 print_endline s;
											 exit 2
										       end
					    in
					      saveRec 1048576 (* 1 Mo *)
